/////////////////////////////////////////////////////////////////////////////////////////////////////////
//   _     ____ ____    _   _  ___  _  _____    _      __________ _  ___  
//  | |   / ___|  _ \  | \ | |/ _ \| |/ /_ _|  / \    |___ /___ // |/ _ \ 
//  | |  | |   | | | | |  \| | | | | ' / | |  / _ \     |_ \ |_ \| | | | |
//  | |__| |___| |_| | | |\  | |_| | . \ | | / ___ \   ___) |__) | | |_| |
//  |_____\____|____/  |_| \_|\___/|_|\_\___/_/   \_\ |____/____/|_|\___/ 
//
//  ALL YOU NEED TO DO GRAPHICS WITH A NOKIA3310 LCD SCREEN AND AN ARDUINO
//
//						CODED BY FRASER ASHWORTH 2009
/////////////////////////////////////////////////////////////////////////////////////////////////////////
#include "LCD_Nokia_3310.h"
#include <avr/pgmspace.h>
#include <avr/io.h>
#include "WConstants.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CTOR
LCD_Nokia_3310::LCD_Nokia_3310()
{
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	INITIALISE
void LCD_Nokia_3310::Initialise(void)
{
  DDRB |= (1<<SPI_CS)|(1<<LCD_DC)|(1<<LCD_RST)|(1<<SPI_MOSI)|(1<<SPI_SCK);
  DDRD |= (1<<LCD_BL);
  
  PORTB &= ~(1<<LCD_RST);
  	
  delayMicroseconds(1);

  PORTB |= (1<<LCD_RST);  
	
	SPCR = 0x51;   // enable SPI master, fosc/16 = 1MH
   
    WriteByte(0x21, true);	
    WriteByte(0xc8, true);	
    WriteByte(0x06, true);	
    WriteByte(0x13, true);	
    WriteByte(0x20, true);	
    
	ClearBuffer();
    RenderBuffer();
	  
    WriteByte(0x0c, true);	

	BackLight(true);
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	TURN BACKLIGHT ON OR OFF
void LCD_Nokia_3310::BackLight(bool bOn)
{
	if(bOn)
		PORTD |= (1<<LCD_BL);  // turn on backlight
	else
		PORTD &=~(1<<LCD_BL);  // turn off backlight
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CLEAR THE BUFFER
void LCD_Nokia_3310::ClearBuffer(void)
{
	for (int iByte = 0; iByte < BYTES; iByte++)
		pBuffer[iByte] = 0;
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
// BLIT BUFFER RAM TO THE SCREEN
void LCD_Nokia_3310::RenderBuffer(void)
{
	SetXY(0,0);
	for (int iByte = 0; iByte < BYTES; iByte++)
		WriteByte(pBuffer[iByte], false);
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
// PLOT A PIXEL IN THE BUFFER
void LCD_Nokia_3310::Pixel(int x, int y, bool bInvert)
{
	//bounds checking:
	if(x < 0)
		return;
	if(x >= WIDTH)
		return;
	if(y < 0)
		return;
	if(y >= HEIGHT)
		return;

	//calculate bitwise manipulation for correct pixel in monochrome buffer:
	int iByte = ((y>>3) * WIDTH) + x;
	unsigned char ucMask = 0x01 << (y%8);
	if(bInvert)
		pBuffer[iByte] &= ~ucMask;
	else
		pBuffer[iByte] |= ucMask;
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
// INLINE SWAP MACRO
int tmpswap;
#define SWAP(a,b) tmpswap = a; a = b; b = tmpswap;
/////////////////////////////////////////////////////////////////////////////////////////////////////////
// DRAW A LINE IN THE BUFFER (optimised Bresenham algorithm)
void LCD_Nokia_3310::Line(int x0, int y0, int x1, int y1, bool bInvert)
{
	int steep = 1;
    int dx = abs(x1 - x0);//delta X
    int dy = abs(y1 - y0);//delta Y
    int sx = ((x1 - x0) > 0) ? 1 : -1;//step positive or negative (1 or -1)
    int sy = ((y1 - y0) > 0) ? 1 : -1;//step positive or negative (1 or -1)

    if (dy > dx)
	{
	    steep = 0;
	    SWAP(x0, y0);
	    SWAP(dx, dy);
	    SWAP(sx, sy);
	}
	
    int e = (dy << 1) - dx;
    for (int i = 0; i < dx; i++)
	{ 
	    if (steep)
			Pixel(x0, y0, bInvert);
		else
			Pixel(y0, x0, bInvert);
		
	    while (e >= 0)
		{
		    y0 += sy;
		    e -= (dx << 1);
		}
	    x0 += sx;
	    e += (dy << 1);
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
// DRAW AN EMPTY RECTANGLE IN THE BUFFER
void LCD_Nokia_3310::Rectangle(int x0, int y0, int x1, int y1, bool bInvert)
{
	Line(x0, y0, x1, y0, bInvert);//N
	Line(x1, y0, x1, y1+1, bInvert);//E
	Line(x0, y1, x1, y1, bInvert);//S
	Line(x0, y0, x0, y1, bInvert);//W
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
// DRAW A FILLED RECTANGLE IN THE BUFFER
void LCD_Nokia_3310::Box(int x0, int y0, int x1, int y1, bool bInvert)
{
	if(x1 > x0)
	{
		for(int x = x0; x <= x1; x++)
			Line(x, y0, x, y1, bInvert);
	}
	else
	{
		for(int x = x0; x >= x1; x--)
			Line(x, y0, x, y1, bInvert);
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
// DRAW A FAST FILLED TRIANGLE IN THE BUFFER
void LCD_Nokia_3310::Triangle(int x0, int y0, int x1, int y1, int x2, int y2, bool bInvert)
{
	//sort triangle points in Y dimension:
	for(int s = 0; s < 2; s++)
	{
		if(y0 > y1)
		{
			SWAP(y0,y1);
			SWAP(x0,x1);
		}
		if(y1 > y2)
		{
			SWAP(y1,y2);
			SWAP(x1,x2);
		}
		if(y0 > y2)
		{
			SWAP(y0,y2);
			SWAP(x0,x2);
		}
	}
	
	//traingle is split into upper and lower half, upper has flat bottom, lower has flat top
	int iTopRows = y1 - y0;
	int iBotRows = y2 - y1;
	int iAllRows = y2 - y0;
	if(iAllRows == 0)
		return;
	
	//float math to work out edge gradients
	float fX1, fX2, fDX1, fDX2;
	
	//Fill top half
	fX1 = x0;
	fDX1 = (x2 - x0) / (float)iAllRows;
	if(iTopRows > 0)
	{
		fX2 = x0;
		fDX2 = (x1 - x0) / (float)iTopRows;
		for(int y = y0; y <= y1; y++)
		{
			Line(fX1, y, fX2, y, bInvert);
			fX1 += fDX1;
			fX2 += fDX2;
		}
	}

	//Fill bottom half
	if(iBotRows > 0)
	{
		fX2 = x1;
		fDX2 = (x2 - x1) / (float)iBotRows;
		for(int y = y1; y <= y2; y++)
		{
			Line(fX1, y, fX2, y, bInvert);
			fX1 += fDX1;
			fX2 += fDX2;
		}
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//LARGE DIGIT NUMBER WITH COMMA SEPERATOR
//	set iInvert to zero to have no effect
void LCD_Nokia_3310::LargeNumber(int x0, int y0, long lNumber, int iInvert)
{
	const int NUM_DIGITS = 8;
	char* pStr = ltoa(lNumber, NUM_DIGITS);
	for(int i = 0; i < NUM_DIGITS; i++)
	{
		LargeDigit(x0, y0, pStr[i]-'0', i==iInvert);
		x0 += 10;
		if(i == 1 || i == 4)
		{	//comma:
			Pixel(x0,y0+9,false);
			Pixel(x0+1,y0+9,false);
			Pixel(x0,y0+10,false);
			Pixel(x0+1,y0+10,false);
			Pixel(x0+1,y0+11,false);
			x0 += 2;
			
		}
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//LARGE DIGIT FONT
//Each character occupies a 10 x 14 grid of screen estate
void LCD_Nokia_3310::LargeDigit(int x0, int y0, int iDigit, bool bInvert)
{
	unsigned char  LARGE_DIGIT_FONT[][10] PROGMEM =
	{
		{ 0x7E, 0xFF, 0xC3, 0xC3, 0xC3, 0xC3, 0xC3, 0xC3, 0xFF, 0x7E },//0
		{ 0x0C, 0x1C, 0x1C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C, 0x0C },//1
		{ 0x7E, 0xFF, 0xC3, 0x03, 0x7F, 0xFE, 0xC0, 0xC0, 0xFF, 0xFF },//2
		{ 0x7E, 0xFF, 0xC3, 0x03, 0x3E, 0x3E, 0x03, 0xC3, 0xFF, 0x7E },//3
		{ 0xC3, 0xC3, 0xC3, 0xC3, 0xFF, 0x7F, 0x03, 0x03, 0x03, 0x03 },//4
		{ 0xFF, 0xFF, 0xC0, 0xC0, 0xFE, 0xFF, 0x03, 0x03, 0xFF, 0xFE },//5
		{ 0x7E, 0xFE, 0xC0, 0xC0, 0xFE, 0xFF, 0xC3, 0xC3, 0xFF, 0x7E },//6
		{ 0xFF, 0xFF, 0x03, 0x03, 0x7F, 0x7F, 0x03, 0x03, 0x03, 0x03 },//7
		{ 0x7E, 0xFF, 0xC3, 0xC3, 0x7E, 0x7E, 0xC3, 0xC3, 0xFF, 0x7E },//8
		{ 0x7E, 0xFF, 0xC3, 0xC3, 0xFF, 0x7F, 0x03, 0x03, 0x7F, 0x7E },//9
	};	
	unsigned char* pFont = (unsigned char*)LARGE_DIGIT_FONT[iDigit%10];
	x0++;
	y0++;
	for(int y = 0; y < 10; y++)
		for(int x = 0; x < 8; x++)
			if(pFont[y] & (0x80>>x))
				Pixel(x+x0, y+y0, bInvert);
			else
				Pixel(x+x0, y+y0, !bInvert);
	Rectangle(x0-1, y0-1, x0+8, y0+10, !bInvert);
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//SMALL TEXT FONT
//Each character occupies a 6 x 8 grid of screen estate
void LCD_Nokia_3310::SmallChar(int x0, int y0, unsigned char ucChar, bool bInvert)
{
	// FONT, pixel matrix rotated 90deg clockwise
	//  1 pixel space at left and bottom
	//  index = ASCII - 32
	unsigned char  SMALL_TEXT_FONT[][6] PROGMEM =
	{
		{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },   // sp
		{ 0x00, 0x00, 0x00, 0x2f, 0x00, 0x00 },   // !
		{ 0x00, 0x00, 0x07, 0x00, 0x07, 0x00 },   // "
		{ 0x00, 0x14, 0x7f, 0x14, 0x7f, 0x14 },   // #
		{ 0x00, 0x24, 0x2a, 0x7f, 0x2a, 0x12 },   // $
		{ 0x00, 0x62, 0x64, 0x08, 0x13, 0x23 },   // %
		{ 0x00, 0x36, 0x49, 0x55, 0x22, 0x50 },   // &
		{ 0x00, 0x00, 0x05, 0x03, 0x00, 0x00 },   // '
		{ 0x00, 0x00, 0x1c, 0x22, 0x41, 0x00 },   // (
		{ 0x00, 0x00, 0x41, 0x22, 0x1c, 0x00 },   // )
		{ 0x00, 0x14, 0x08, 0x3E, 0x08, 0x14 },   // *
		{ 0x00, 0x08, 0x08, 0x3E, 0x08, 0x08 },   // +
		{ 0x00, 0x00, 0x00, 0xA0, 0x60, 0x00 },   // ,
		{ 0x00, 0x08, 0x08, 0x08, 0x08, 0x08 },   // -
		{ 0x00, 0x00, 0x60, 0x60, 0x00, 0x00 },   // .
		{ 0x00, 0x20, 0x10, 0x08, 0x04, 0x02 },   // /
		{ 0x00, 0x3E, 0x51, 0x49, 0x45, 0x3E },   // 0
		{ 0x00, 0x00, 0x42, 0x7F, 0x40, 0x00 },   // 1
		{ 0x00, 0x42, 0x61, 0x51, 0x49, 0x46 },   // 2
		{ 0x00, 0x21, 0x41, 0x45, 0x4B, 0x31 },   // 3
		{ 0x00, 0x18, 0x14, 0x12, 0x7F, 0x10 },   // 4
		{ 0x00, 0x27, 0x45, 0x45, 0x45, 0x39 },   // 5
		{ 0x00, 0x3C, 0x4A, 0x49, 0x49, 0x30 },   // 6
		{ 0x00, 0x01, 0x71, 0x09, 0x05, 0x03 },   // 7
		{ 0x00, 0x36, 0x49, 0x49, 0x49, 0x36 },   // 8
		{ 0x00, 0x06, 0x49, 0x49, 0x29, 0x1E },   // 9
		{ 0x00, 0x00, 0x36, 0x36, 0x00, 0x00 },   // :
		{ 0x00, 0x00, 0x56, 0x36, 0x00, 0x00 },   // ;
		{ 0x00, 0x08, 0x14, 0x22, 0x41, 0x00 },   // <
		{ 0x00, 0x14, 0x14, 0x14, 0x14, 0x14 },   // =
		{ 0x00, 0x00, 0x41, 0x22, 0x14, 0x08 },   // >
		{ 0x00, 0x02, 0x01, 0x51, 0x09, 0x06 },   // ?
		{ 0x00, 0x32, 0x49, 0x59, 0x51, 0x3E },   // @
		{ 0x00, 0x7C, 0x12, 0x11, 0x12, 0x7C },   // A
		{ 0x00, 0x7F, 0x49, 0x49, 0x49, 0x36 },   // B
		{ 0x00, 0x3E, 0x41, 0x41, 0x41, 0x22 },   // C
		{ 0x00, 0x7F, 0x41, 0x41, 0x22, 0x1C },   // D
		{ 0x00, 0x7F, 0x49, 0x49, 0x49, 0x41 },   // E
		{ 0x00, 0x7F, 0x09, 0x09, 0x09, 0x01 },   // F
		{ 0x00, 0x3E, 0x41, 0x49, 0x49, 0x7A },   // G
		{ 0x00, 0x7F, 0x08, 0x08, 0x08, 0x7F },   // H
		{ 0x00, 0x00, 0x41, 0x7F, 0x41, 0x00 },   // I
		{ 0x00, 0x20, 0x40, 0x41, 0x3F, 0x01 },   // J
		{ 0x00, 0x7F, 0x08, 0x14, 0x22, 0x41 },   // K
		{ 0x00, 0x7F, 0x40, 0x40, 0x40, 0x40 },   // L
		{ 0x00, 0x7F, 0x02, 0x0C, 0x02, 0x7F },   // M
		{ 0x00, 0x7F, 0x04, 0x08, 0x10, 0x7F },   // N
		{ 0x00, 0x3E, 0x41, 0x41, 0x41, 0x3E },   // O
		{ 0x00, 0x7F, 0x09, 0x09, 0x09, 0x06 },   // P
		{ 0x00, 0x3E, 0x41, 0x51, 0x21, 0x5E },   // Q
		{ 0x00, 0x7F, 0x09, 0x19, 0x29, 0x46 },   // R
		{ 0x00, 0x46, 0x49, 0x49, 0x49, 0x31 },   // S
		{ 0x00, 0x01, 0x01, 0x7F, 0x01, 0x01 },   // T
		{ 0x00, 0x3F, 0x40, 0x40, 0x40, 0x3F },   // U
		{ 0x00, 0x1F, 0x20, 0x40, 0x20, 0x1F },   // V
		{ 0x00, 0x3F, 0x40, 0x38, 0x40, 0x3F },   // W
		{ 0x00, 0x63, 0x14, 0x08, 0x14, 0x63 },   // X
		{ 0x00, 0x07, 0x08, 0x70, 0x08, 0x07 },   // Y
		{ 0x00, 0x61, 0x51, 0x49, 0x45, 0x43 },   // Z
/*		{ 0x00, 0x00, 0x7F, 0x41, 0x41, 0x00 },   // [
		{ 0x00, 0x55, 0x2A, 0x55, 0x2A, 0x55 },   // 55
		{ 0x00, 0x00, 0x41, 0x41, 0x7F, 0x00 },   // ]
		{ 0x00, 0x04, 0x02, 0x01, 0x02, 0x04 },   // ^
		{ 0x00, 0x40, 0x40, 0x40, 0x40, 0x40 },   // _
		{ 0x00, 0x00, 0x01, 0x02, 0x04, 0x00 },   // '
		{ 0x00, 0x20, 0x54, 0x54, 0x54, 0x78 },   // a
		{ 0x00, 0x7F, 0x48, 0x44, 0x44, 0x38 },   // b
		{ 0x00, 0x38, 0x44, 0x44, 0x44, 0x20 },   // c
		{ 0x00, 0x38, 0x44, 0x44, 0x48, 0x7F },   // d
		{ 0x00, 0x38, 0x54, 0x54, 0x54, 0x18 },   // e
		{ 0x00, 0x08, 0x7E, 0x09, 0x01, 0x02 },   // f
		{ 0x00, 0x18, 0xA4, 0xA4, 0xA4, 0x7C },   // g
		{ 0x00, 0x7F, 0x08, 0x04, 0x04, 0x78 },   // h
		{ 0x00, 0x00, 0x44, 0x7D, 0x40, 0x00 },   // i
		{ 0x00, 0x40, 0x80, 0x84, 0x7D, 0x00 },   // j
		{ 0x00, 0x7F, 0x10, 0x28, 0x44, 0x00 },   // k
		{ 0x00, 0x00, 0x41, 0x7F, 0x40, 0x00 },   // l
		{ 0x00, 0x7C, 0x04, 0x18, 0x04, 0x78 },   // m
		{ 0x00, 0x7C, 0x08, 0x04, 0x04, 0x78 },   // n
		{ 0x00, 0x38, 0x44, 0x44, 0x44, 0x38 },   // o
		{ 0x00, 0xFC, 0x24, 0x24, 0x24, 0x18 },   // p
		{ 0x00, 0x18, 0x24, 0x24, 0x18, 0xFC },   // q
		{ 0x00, 0x7C, 0x08, 0x04, 0x04, 0x08 },   // r
		{ 0x00, 0x48, 0x54, 0x54, 0x54, 0x20 },   // s
		{ 0x00, 0x04, 0x3F, 0x44, 0x40, 0x20 },   // t
		{ 0x00, 0x3C, 0x40, 0x40, 0x20, 0x7C },   // u
		{ 0x00, 0x1C, 0x20, 0x40, 0x20, 0x1C },   // v
		{ 0x00, 0x3C, 0x40, 0x30, 0x40, 0x3C },   // w
		{ 0x00, 0x44, 0x28, 0x10, 0x28, 0x44 },   // x
		{ 0x00, 0x1C, 0xA0, 0xA0, 0xA0, 0x7C },   // y
		{ 0x00, 0x44, 0x64, 0x54, 0x4C, 0x44 },   // z
		{ 0x00, 0x00, 0x06, 0x09, 0x09, 0x06 }    // horiz lines*/
	};
	unsigned char* pFont = (unsigned char*)SMALL_TEXT_FONT[ucChar-32];
	for (unsigned char x = 0; x<6; x++)
	{
		for (unsigned char y = 0; y<8; y++)
			if(pFont[x] & (0x01<<y))
				Pixel(x+x0, y+y0, bInvert);
			else
				Pixel(x+x0, y+y0, !bInvert);
		Pixel(x+x0, y0-1, !bInvert);
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	DRAW A STRING
//  Set iInvert to 100 to make it all get inverted
void LCD_Nokia_3310::SmallString(int x0, int y0, char* pStr, int iInvert)
{
	while(*pStr)
	{
		SmallChar(x0, y0, *pStr++, iInvert==0 || iInvert==100);
		if(iInvert < 100)
			iInvert--;
		x0 += 6;
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	DRAW A NUMBER
//  Set iInvert to 100 to make it all get inverted
void LCD_Nokia_3310::SmallNumber(int x0, int y0, long lNumber, int iDigits, int iInvert)
{
	char* pStr = ltoa(lNumber, iDigits);
	for(int i = 0; i < iDigits; i++)
	{
		SmallChar(x0, y0, pStr[i], i==iInvert || iInvert==100);
		x0 += 6;
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	DRAW A BITMAP
void LCD_Nokia_3310::Bitmap(int x, int y, int w, int h, unsigned char* pBmp)
{
	int iPix = 0;
	int iByte = 0;
	for(int iY = y; iY < (y+h); iY++)
	{
		for(int iX = x; iX < (x+w); iX++)
		{
			if(pBmp[iByte] & (0x80>>iPix))
				Pixel(iX, iY, false);

			if(++iPix >= 8)
			{
				iPix = 0;
				iByte++;
			}
		}
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	 ____  ____  _____     ___  _____ _____
//	|  _ \|  _ \|_ _\ \   / / \|_   _| ____|
//	| |_) | |_) || | \ \ / / _ \ | | |  _|
//	|  __/|  _ < | |  \ V / ___ \| | | |___
//	|_|   |_| \_\___|  \_/_/   \_\_| |_____|
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////
// SET CURSOR POSITION	
void LCD_Nokia_3310::SetXY(unsigned char x, unsigned char y)
{
    WriteByte(0x40 | y, true);		// column
    WriteByte(0x80 | x, true);      // row
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//	SEND A BYTE TO THE SCREEN USING SPI INTERFACE
void LCD_Nokia_3310::WriteByte(unsigned char byte, bool bControl)
{
    PORTB &= ~(1<<SPI_CS);
    
	if (bControl)
		PORTB &= ~(1<<LCD_DC);
    else
		PORTB |= (1<<LCD_DC);
	
	SPDR = byte;			
	while ((SPSR & 0x80) == 0);   
	
	PORTB |= (1<<SPI_CS);	
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Convert number to ascii string
char* LCD_Nokia_3310::ltoa(long lNum, int iNumDigits)
{
	const int MAX_DIGITS = 10;
	static char pChar[MAX_DIGITS + 1];
	
	unsigned long ulDiv = 1;
	for(int i = 1; i < iNumDigits; i++)
		ulDiv *= 10;
	
	if (iNumDigits > MAX_DIGITS)
		iNumDigits = MAX_DIGITS;
	
	for(int i = 0; i < iNumDigits; i++)
    {
		char digit = ((lNum/ulDiv) % 10);
		pChar[i] = digit + '0';
		ulDiv /= 10;
    }
    pChar[iNumDigits] = 0;//terminator
    return pChar;
}